# Pantry, command-line nutrient analysis. Version 11
# Copyright 2007 Omari Norman. 
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE. 
# 
# Pantry's "built-in" XML validator is xmlproc.
# xmlproc is Copyright 1998-2000 by Lars Marius Garshol, Oslo, Norway. 
# 
# All Rights Reserved 
# 
# Permission to use, copy, modify, and distribute this software
# and its documentation for any purpose and without fee is hereby
# granted, provided that the above copyright notice appear in all
# copies and that modified copies are clearly marked as such. 
# 
# LARS MARIUS GARSHOL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
# SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS, IN NO EVENT SHALL LARS MARIUS GARSHOL BE LIABLE FOR
# ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
# IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
# ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
# THIS SOFTWARE. 
""" Main Pantry command-line interface.
    
These are meant to be called using trivial Python scripts that
may be installed in the user's PATH. main is the big kahuna.
addTo relies on main, so changes to the main
command-line options might break addTo. 

Both main and addTo take a single argv argument. argv must be
the entire argv of the calling script (including argv[0].)"""

from __future__ import division

def main(argv):
    from optparse import OptionParser
    import errno
    import os
    import sys
    import textwrap
    
    from pantry import unicodeHelper
    from pantry import version
    from pantry import xmlCheck
    from pantry.lib import cabinets
    from pantry.lib import foods
    import config
    import reports

    prog = os.path.basename(argv[0])
    usage = 'usage: %prog [options] FILE...'

    parser = OptionParser(prog = prog, usage = usage)

    # search options
    parser.add_option('-i', '--ignore-case', action='store_const',
            const=cabinets.IGNORECASE,
            dest='ignoreCase', 
            help="make searches case insensitive")
    parser.add_option('-x', '--exact-match', action='store_const',
            const=cabinets.EXACTMATCH,
            dest='exactMatch',
            help='traits must exactly match search term specified')
    parser.add_option("-n", "--name", action="store",
            type="string", dest="name", 
            metavar='regexp',
            help='select foods with matching name')
    parser.add_option('-g', '--group', action="store",
            type='string', dest='group', 
            metavar='regexp',
            help='select foods with matching group')
    parser.add_option("-d", "--date", action="store",
            type="string", dest='date', 
            metavar='regexp',
            help='select foods with matching date')
    parser.add_option('-m', '--meal', action='store',
            type='string', dest='meal', 
            metavar='regexp',
            help='select foods with matching meal')
    parser.add_option('-u', '--unit', action='store',
            type='string', dest='unit', 
            metavar='regexp',
            help='select foods with matching unit')
    parser.add_option('-q', '--qty', action='store',
            type='string', dest='qty', 
            metavar='number',
            help='select foods with matching qty')
    parser.add_option('-c', '--comment', action='store',
            type='string', dest='comment',
            metavar='regexp',
            help='select foods with matching comment')
    parser.add_option('-o', '--order', action='store',
            type='string', dest='order',
            metavar='regexp',
            help='select foods with matching order')
    parser.add_option('-p', '--print', action='store',
            type='string', dest='report', metavar = 'REPORT',
            help='print REPORT')
    parser.add_option('-l', '--nutrient-list', action='store',
            type='string', dest='nutList', metavar='NUTRIENT-LIST',
            help='use NUTRIENT-LIST when printing report')
    parser.add_option('-a', '--add', action='append',
            type='string', dest='addFiles', metavar='FILE',
            help='add results to FILE')
    parser.add_option('-s', '--sort', action='store', type='string',
            dest='sort', metavar='traits',
            help='sort foods in report by these traits')
    parser.add_option('--force-valid', action='store_const',
            const = cabinets.ALWAYS, dest='validate',
            help='validate all XML files, even if above 300 kilobytes')
    parser.add_option('--skip-valid', action='store_const',
            const = cabinets.NEVER, dest = 'validate',
            help = 'do not validate any XML files')
    
    # change options
    parser.add_option('-N', '--c-name', action='store',
            type='string', dest='c_name', metavar='string',
            help='change name to string')
    parser.add_option('-G', '--c-group', action='store',
            type='string', dest='c_group', metavar='string',
            help='change group to string')
    parser.add_option('-Q', '--c-qty', action='store',
            type='string', dest='c_qty', metavar='number',
            help='change quantity to number')
    parser.add_option('-U', '--c-unit', action='store',
            type='string', dest='c_unit', metavar='regexp',
            help='change unit to one matching regexp')
    parser.add_option('-D', '--c-date', action='store',
            type='string', dest='c_date', metavar='string',
            help='change date to string')
    parser.add_option('-M', '--c-meal', action='store',
            type='string', dest='c_meal', metavar='string',
            help='change meal to string')
    parser.add_option('-C', '--c-comment', action='store',
            type='string', dest='c_comment', metavar='string',
            help='change comment to string')
    parser.add_option('-O', '--c-order', action='store',
            type='string', dest='c_order', metavar='string',
            help='change order to string')
    parser.add_option('-A', '--auto-order', action='store_true',
            dest='autoOrder', 
            help='automatically increment order trait')
    parser.add_option('--by-nut', action='store',
            type='string', dest='c_by_nut', metavar='nut-name qty',
            nargs=2,
            help='set quantity of food so amount of nut-name = qty')
    parser.add_option('-K', '--by-cal', action='store',
            type='string', dest='c_by_cal', metavar='calories',
            help='set quantity of food so calories equals calories')
    parser.add_option('--edit', action='store_true',
            dest='edit', help='Edit foods in place') 
    parser.add_option('--delete', action='store_true', 
            dest='delete', help='delete matching foods')
    parser.add_option('-R', '--refuse', action='store_true',
            dest='useRefuse', 
            help='reduce qty by percent refuse')

    # dump options
    parser.add_option('--dump', action='store', type='string',
            dest='dump', metavar='what-to-dump',
            help='print internal Pantry variables (details in man page)')
    parser.add_option('-v', '--version', action='store_const',
            const='version', dest='dump',
            help='Print version and copyright information and exit')

    parser.set_defaults(ignoreCase=0, exactMatch=0,
            addFiles = [], sort='', autoOrder = False,
            validate = cabinets.SMALL)

    (options, posArgs) = parser.parse_args(argv[1:])

    # convert to unicode
    options = unicodeHelper.convertOptions(options)
    posArgs = unicodeHelper.convertList(posArgs)

    stderr = unicodeHelper.puni(sys.stderr)
    errors = unicodeHelper.ErrorPrinter(stderr, prog)
    stdout = unicodeHelper.puni(sys.stdout)
    searchOpts = options.ignoreCase | options.exactMatch

    try:
        configVars = config.ConfigVars()
    except config.Error, e:
        errors(u'Problem loading Pantry RC file:', e, 
                type = unicodeHelper.ERROR)

    if options.c_by_cal:
        options.c_by_nut = [u'Calories', options.c_by_cal]

    if options.dump:
        for dumpName, whatToPrint in (('version', version.COPYRIGHT),
            ('recipesDTD', textwrap.dedent(cabinets.Recipes.DTD)),
            ('foodsDTD', textwrap.dedent(cabinets.Xml.DTD)),
            ('rcDTD', textwrap.dedent(configVars.DTD)),
            ('config', str(configVars))):
            if options.dump == dumpName:
                stdout(whatToPrint + '\n')
                sys.exit()
        errors(u'unrecognized argument for dump option.',
                type = unicodeHelper.ERROR)

    if len(posArgs) < 1:
        errors('you did not specify any files to search.')

    # do we need nutrients?
    needNuts = False
    if options.report:
        reporter = reports.Reports(options.report, options.sort,
                configVars.sortOrders, prog)
        needNuts = reporter.getNeedNuts()
    if options.edit or options.addFiles or options.c_by_nut:
        needNuts = True
    results = cabinets.Temp()

    for filename in posArgs:
        # search files
        try:
            target = cabinets.getCabinet(filename,
                    pathList = configVars.pathList,
                    validate = options.validate, create = False)
        except cabinets.CabinetOpenError, e:
            errors('could not open file %s' % filename, e)
            continue

        searchDict = dict([(trait, getattr(options, trait))
                    for trait in ('name', 'group', 'date', 'meal',
                        'qty', 'unit', 'comment', 'order')
                    if getattr(options, trait) != None])
        keyList = target.search(searchDict, searchOpts)

        for key in keyList:
            # change foods; fill results instance
            newFood = target.getFood(key, needNuts).copy()
            newTraits = [(trait, getattr(options, 'c_' + trait))
                    for trait in ('name', 'group', 'date', 'meal',
                        'qty', 'comment', 'order')
                    if getattr(options, 'c_' + trait) != None]
            try:
                newFood.setTraits(newTraits)
            except foods.SetTraitError, e:
                errors('food %s will not be copied into the buffer.' %
                        newFood.traits['name'], e)
                continue

            if options.c_unit != None:
                try:
                    newFood.setUnit(options.c_unit, searchOpts)
                except foods.SetUnitError, e:
                    errors('food %s will not be copied into the buffer.' %
                            newFood.traits['name'], e)
                    continue

            if options.c_by_nut:
                newQty = newFood.getQtyByNut(options.c_by_nut[0],
                        options.c_by_nut[1], searchOpts)
                newFood.traits['qty'] = newQty

            if options.useRefuse:
                newQty = newFood.qtyMinusRefuse
                newFood.traits['qty'] = newQty

            results.addFood(newFood)
            if options.delete or options.edit:
                try:
                    del target[key]
                except cabinets.ReadOnlyError:
                    emsg = (u'file %s is read only;' % filename
                            + u'food %s will not be edited or deleted.' %
                            target[key].traits['name'])
                    errors(emsg)
                    continue
            if options.edit:
                target.addFood(newFood)
        try:
            target.save()
        except AttributeError:
            pass
        del target

    if options.report:
        if options.nutList:
            nutList = configVars.nutLists[options.nutList]
        else:
            nutList = configVars.nutLists[configVars.defaultNutList]
        reporter.printReports(results, nutList)

    for addFile in options.addFiles:
        try:
            saveTarget = cabinets.getCabinet(addFile, 
                    pathList = configVars.pathList,
                    validate = options.validate, create = True)
        except cabinets.CabinetOpenError, e:
            errors(u'could not open for --add file %s' % filename, e)
            continue

        for food in results.values():
            try:
                saveTarget.addFood(food, options.autoOrder)
            except cabinets.ReadOnlyError:
                errors(u'file %s is read only; no foods' % addFile
                        + u'will be added to it.')
                break
        saveTarget.save()

def addTo(argv):
    from optparse import OptionParser
    from pantry import unicodeHelper
    import datetime
    import os
    import sys

    import config

    configVars = config.ConfigVars()
    stderr = unicodeHelper.puni(sys.stderr)
    prog = os.path.basename(argv[0])
    errors = unicodeHelper.ErrorPrinter(stderr, prog)
    if prog == 'pantry-addTo':
        errors("pantry-addTo: error: don't invoke this program"
                " directly. Please consult"
                " the Pantry Manual for details.",
                type = unicodeHelper.ERROR)

    usage = 'usage: %prog FOOD-NAME [QTY] [UNITS]'
    parser = OptionParser(prog = prog, usage = usage)
    parser.add_option('-g', '--group', action='store_true',
            dest='group', help='search by group name rather than FOOD-NAME',
            default = False)
    parser.add_option('-R', '--refuse', action='store_true',
            dest='useRefuse', help='reduce food qty by percent refuse',
            default = False)

    (options, posArgs) = parser.parse_args(argv[1:])

    meals = configVars.symlinks
    try:
        meal = meals[prog]
    except KeyError:
        errors('meal corresponding to this script name'
                ' is not defined.', unicodeHelper.ERROR)

    if options.group:
        searchCmd = '-ig'
    else:
        searchCmd = '-in'

    args = [ prog ]
    if len(posArgs) == 1:
        args.extend( [searchCmd, posArgs[0]] )
    elif len(posArgs) == 2:
        args.extend( [searchCmd, posArgs[0], '-Q', posArgs[1]]) 
    elif len(posArgs) == 3:
        args.extend( [searchCmd, posArgs[0], '-Q', posArgs[1], 
            '-U', posArgs[2]] )
    else:
        errors('invalid number of arguments entered.', unicodeHelper.ERROR)

    today = str(datetime.date.today())
    args.extend(['-D', today, '-M', meal])
    args.extend(['-p', 'traits-blank'])
    args.append('--auto-order')
    if options.useRefuse:
        args.append('--refuse')

    destination = configVars.destination
    source = configVars.source
    args.extend(['-a', destination, source])

    main(args)
