# Pantry, command-line nutrient analysis. Version 11
# Copyright 2007 Omari Norman. 
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE. 
# 
# Pantry's "built-in" XML validator is xmlproc.
# xmlproc is Copyright 1998-2000 by Lars Marius Garshol, Oslo, Norway. 
# 
# All Rights Reserved 
# 
# Permission to use, copy, modify, and distribute this software
# and its documentation for any purpose and without fee is hereby
# granted, provided that the above copyright notice appear in all
# copies and that modified copies are clearly marked as such. 
# 
# LARS MARIUS GARSHOL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
# SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS, IN NO EVENT SHALL LARS MARIUS GARSHOL BE LIABLE FOR
# ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
# IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
# ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
# THIS SOFTWARE. 
from __future__ import division
import sys
from pantry import unicodeHelper
stdout = unicodeHelper.puni()
stderr = unicodeHelper.puni(sys.stderr)

class Reports(object):
    """ Print out food reports. """
    
    SUMMARY_REPORTS = ['sum', 'groups', 'list']
    NUT_REPORTS = ['sum', 'nuts', 'xml']

    def __init__(self, reports, sortFields = '', sortOrders = [], prog = None):
        from pantry.lib import foods
        reports = reports.split('-')
        self.foodReports = []
        self.summaryReports = []
        self.sortOrders = sortOrders
        self.prog = prog
        availReports = vars(Reports).keys()

        for report in reports:
            possRpts = [ rptName for rptName in availReports
                    if rptName.startswith(report) ]
            if len(possRpts) != 1:
                err = (("%s: warning: report name %s does not exist" %
                        (prog, report))
                        + " or is ambiguous\n")
                stderr(err)
            else:
                if possRpts[0] in self.SUMMARY_REPORTS:
                    self.summaryReports.append(possRpts[0])
                else:
                    self.foodReports.append(possRpts[0])

        self.sortList = []
        for char in sortFields:
            for traitName in foods.TRAITSBYNAME:
                found = False
                if traitName.startswith(char):
                    self.sortList.append(traitName)
                    found = True
                    break
                elif traitName.startswith(char.lower()):
                    self.sortList.append('-' + traitName)
                    found = True
                    break
            if not found:
                err = ("%s: warning: sort trait %s not recognized\n" %
                        (prog, char))
                stderr(err)

        self.needNuts = False
        for report in self.foodReports + self.summaryReports:
            if report in self.NUT_REPORTS:
                self.needNuts = True
                return

    def printReports(self, unsortedFoods, nutList):
        from pantry.lib import foods

        if self.foodReports:
            if 'nuts' in self.foodReports:
                totalNuts = unsortedFoods.nuts
            if self.sortList:
                foodsToPrint = sorted(unsortedFoods.values(), 
                        foods.CmpFoods(self.sortList, self.sortOrders))
            else:
                foodsToPrint = unsortedFoods.values()

            for food in foodsToPrint:
                for report in self.foodReports:
                    if report == 'nuts':
                        getattr(self, report)(food, nutList, totalNuts)
                    else:
                        getattr(self, report)(food, nutList)

        for report in self.summaryReports:
            getattr(self, report)(unsortedFoods, nutList)

    def names(self, food, nutList):
        stdout(food.traits['name'] + '\n')

    def tpaste(self, food, nutList):
        stdout(food.getTxtPaste())

    def paste(self, food, nutList):
        OPTEXACT='-x'
        OPTNAME='-n'
        OPTCHANGEUNIT='-U'

        name = food.traits['name']
        name = "'%s'" % name.replace("'", "'\\''")
        availUnits = food.units
        availUnits = [ "'%s'" % unit.replace("'", "'\\''")
                for unit in availUnits ]

        for availUnit in availUnits:
            stdout(' '.join([OPTEXACT, OPTNAME, name, 
                OPTCHANGEUNIT, availUnit, '\n']))

    def traits(self, food, nutList):
        from pantry import unicodeHelper

        traits = food.traits

        stdout(traits['name'] + '\n')
        if traits['group']:
            stdout('Group: %s\n' % traits['group'])

        atStartOfLine = True
        for traitName in ['date', 'meal', 'order']:
            if traits[traitName]:
                stdout('%s: %s ' % 
                        (traitName.capitalize(), traits[traitName]))
                atStartOfLine = False
        if not atStartOfLine:
            stdout('\n',)

        if traits['comment']:
            stdout('Comment: %s\n' % traits['comment'])

        if unicodeHelper.convertUni(traits['pctRefuse']):
            stdout(('Refuse: %s percent %s\n' %
                    (traits['pctRefuse'], traits['refDesc'])))

        stdout(('%s %s (%.0fg)\n' 
                % (traits['qty'], traits['unit'], food.grams)))

    def measures(self, food, nutList):
        units = food.arbUnits
        for unit, grams in units.items():
            stdout('   %s (%.0fg)\n' % (unit, grams))

    def units(self, food, nutList):
        units = food.arbUnits
        for unit in units:
            stdout('   %s\n' % unit)

    def debug(self, food, nutList):
        stdout(str(food) + '\n')

    def list(self, foods, nutList):
        """ Prints nutList in nice format. """
        from pantry.lib import foods

        format = "%-25s %-14s %-6s\n"
        headerLabels = ('Nutrient Name', 'Units', 'Goal')
        headerString = format % headerLabels
        stdout(headerString)
        stdout("-" * len(headerString) + '\n')

        for nutName, goal in nutList:
            nutUnits = foods.NUTSBYNAME[nutName][1]
            stdout(format % (nutName, nutUnits, goal))

    def nuts(self, food, nutList, totalNuts = None):
        """ Prints nutrient report for food.

        food only needs to have a nuts method
        (therefore, it need not be a food; it can be, for example,
        a File* instance.)

        If totalNuts is set, also prints a column on the far right
        showing each nutrient's percent of the total. 
        totalNuts is a Nuts instance. """
        
        from pantry.lib import foods

        nuts = food.nuts
        headerFormat = "%-25s %-14s %-6s"
        headerLabels = ['Nutrient', 'Amount', '%G' ]
        if totalNuts:
            headerFormat += ' %-6s'
            headerLabels.append('%TOT')
        headerString = headerFormat % tuple(headerLabels)
        stdout(headerString + '\n')
        stdout("-" * len(headerString) + '\n')

        for nutName, goal in nutList:
            nutUnits = foods.NUTSBYNAME[nutName][1]

            # find pctTotal
            try:
                amtTotal = nuts[(nutName, nutUnits)]
            except KeyError:
                # nuts does not contain this nutrient that was listed in
                # nutList. That doesn't mean it is zero; it just means
                # we don't know what the nutrient content is for this
                # nutrient.
                amtTotal = '??'
                pctGoal = '??'
            else:
                try:
                    pctGoal = amtTotal / goal * 100
                except ZeroDivisionError:
                    # Goal for this nutrient is zero. amtTotal is
                    # almost certainly more than zero, which is
                    # infinitely more (percentage wise) than zero, so
                    # print ++.
                    pctGoal = '++'

                except TypeError:
                    # Goal is None.
                    pctGoal = 'NG'
                else:
                    pctGoal = str(int(round(pctGoal)))
                
            if totalNuts:
                nutUnits = foods.NUTSBYNAME[nutName][1]
                try:
                    total = totalNuts[(nutName, nutUnits)]
                except KeyError:
                    # Total does not exist of this nutrient. How to do a
                    # percentage of an unknown quantity?
                    pctTotal = 'NT'
                else:
                    try:
                        pctTotal = amtTotal / total * 100
                    except TypeError:
                        # amtTotal is a string, like '??'
                        pctTotal = '??'
                    except ZeroDivisionError:
                        # Total is zero
                        if amtTotal > 0:
                            pctTotal = '++'
                        elif amtTotal < 0:
                            pctTotal = '--'
                        else:
                            pctTotal = 'NA'
                    else:
                        pctTotal = str(int(round(pctTotal)))

            lineFormat = "%-25s %-4s %-10s %-6s"
            try:
                amtTotal = str(int(round(amtTotal)))
            except TypeError:
                # amtTotal is a string. Just leave it alone.
                pass

            nutUnits = foods.NUTSBYNAME[nutName][1]
            lineData = [ nutName, amtTotal, nutUnits, pctGoal ]

            if totalNuts:
                lineFormat += ' %-6s\n'
                lineData.append(pctTotal)
            else:
                lineFormat += '\n'

            stdout(lineFormat % tuple(lineData))

    def blank(self, food, nutList):
        stdout('\n')

    def sum(self, foods, nutList):
        stdout("SUM:\n")
        self.nuts(foods, nutList)

    def groups(self, foods, nutList):
        """ Prints group names and number of foods in each group. """
        groupDict = foods.groups
        stdout('  %-35s %7s\n' % ('Group', 'No. of Foods'))
        stdout('-' * 60 + '\n')
        for group in sorted(groupDict):
            stdout('  %-35s %7u\n' % (group, len(groupDict[group])))
        stdout('-' * 60 + '\n')
        stdout('  %-35s %7u foods total\n' % ('', len(foods)))
        stdout('  %-35s %7u groups total\n' % ('', len(groupDict)))

    def getNeedNuts(self):
        return self.needNuts
