# Pantry, command-line nutrient analysis. Version 11
# Copyright 2007 Omari Norman. 
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE. 
# 
# Pantry's "built-in" XML validator is xmlproc.
# xmlproc is Copyright 1998-2000 by Lars Marius Garshol, Oslo, Norway. 
# 
# All Rights Reserved 
# 
# Permission to use, copy, modify, and distribute this software
# and its documentation for any purpose and without fee is hereby
# granted, provided that the above copyright notice appear in all
# copies and that modified copies are clearly marked as such. 
# 
# LARS MARIUS GARSHOL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
# SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS, IN NO EVENT SHALL LARS MARIUS GARSHOL BE LIABLE FOR
# ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
# IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
# ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
# THIS SOFTWARE. 
""" Objects that contain foods.

Foods are kept together in Cabinet instances. Each of these classes
is based off Base, which is a dictionary mixin. The Cabinet
classes vary based on the "back end": that is, how the data is
stored. Shelve is the fastest class, as it stores its data in a
Python shelf. Temp is for foods that are used only within the
program: they aren't stored at all (except in RAM.) Zip opens
the zip files that come from USDA. Currently Zip instances
are read-only, and it's unlikely this will ever change.

Xml instances read foods in from an XML file. Recipes read
*recipes* in from an XML file. Recipes are foods made up of other
foods. Recipes instances are read-only. 

All Cabinet instances are mapping objects, so they need keys. The food's
hash is ideal for this; it is an sha1 hash of all the traits of the food.
To get it use the food's hash property.

Ordinarily to read in an existing Cabinet instance, just use
getCabinet. It figures out what sort of instance you want and it picks
the right class and returns it for you. If you want to create a
specific kind of File instance, you can instantiate the exact one
you want. """

from __future__ import division
import UserDict

# validation options. ALWAYS: always validates underlying XML, no
# matter how big the file is. This can make the validator hang with
# huge files. NEVER: never validates; assumes XML is valid. SMALL:
# validates files that are smaller than MAXSIZE (as given in
# xmlCheck); files larger than MAXSIZE raise an exception.

ALWAYS = 1
NEVER = 2
SMALL = 3

# search options. Combine these with a bitwise OR.
IGNORECASE = 1
EXACTMATCH = 2

class Base(UserDict.DictMixin):
    """ Base class for objects that hold several Food objects. 

    Subclassed from this class are classes that are specific to
    how the foods are stored; use those.

    Subclasses must define a self.allTraits dictionary. The keys
    are the sha1 hashes for the foods; the values are the 
    traits instances for each food.  """

    def __delitem__(self, key):
        del self.allTraits[key]
        del self.allNuts[key]
        del self.allArbUnits[key]
        self.changed = True

    def __getitem__(self, key):
        return self.getFood(key, True)

    def __str__(self):
        return 'File instance holding %u foods.' % len(self)

    def keys(self):
        return self.allTraits.keys()

    def _initFood(self, food):
        # check for hash collisions
        newHash = food.hash
        if newHash in self.allTraits:
            food = food.copy()
            mutator = 0
            originalComment = food.traits['comment']
            while newHash in self.allTraits:
                mutator += 1
                food.traits['comment'] = (u'%s (Copy %s)' %
                        (originalComment, mutator))
                newHash = food.hash

        (self.allTraits[newHash], self.allNuts[newHash],
            self.allArbUnits[newHash], narf1) = food.data
        return food

    def addFood(self, food, incrementOrder = False):
        """ Adds a food to this file.

        Better than setitem, because it checks for dupes, and the
        caller need not worry about getting a hash. 
        
        If incrementOrder is True, addFood will search the foods 
        currently in this file for other foods with identical values
        for both the date and meal traits. The result will be sorted
        and if the highest food's order trait matches '^[0-9]{4}$',
        then addFood will take the first three digits, strip any leading
        zeroes, and increment by one. A trailing zero is then
        added, and a leading zero is added if necessary to make the
        result four digits. 
        
        If there are no foods with identical date and meal traits,
        then the new food's order trait is set to 0010.

        The new food's order trait is set to the
        result. If the new food had an order trait, it is lost."""
        if not self.writeable:
            raise ReadOnlyError

        if incrementOrder:
            import re
            import copy 
            import foods
            food = food.copy()
            traits = food.traits
            keys = self.search(
                    {'date':traits['date'],
                    'meal':traits['meal']},
                    EXACTMATCH)
            
            matchingFoods = [ self[key] for key in keys ]
            if matchingFoods:
                matchingFoods.sort(foods.CmpFoods([ 'order' ], []))
                topOrder = matchingFoods[-1].traits['order']
                if re.search(r'^[0-9]{4}$', topOrder):
                    newTopInt = int(topOrder[:3].lstrip('0')) + 1
                    newTopStr = u"%03u0" % newTopInt
                    food.traits['order'] = newTopStr
            else:
                food.traits['order'] = u'0010'

        newFood = self._initFood(food)
        self.changed = True
        return newFood

    def getFood(self, key, needNuts = True):
        """ Get a food from this file.

        Returns a food with nutrients if needNuts is True;
        otherwise, return a food without nutrients. """
        import foods

        arbUnits = self.allArbUnits[key]
        traits = self.allTraits[key]
        if needNuts:
            nuts = self.allNuts[key]
            result = foods.Food(traits, nuts, arbUnits, 
                    origFile = self.filename)
        else:
            result = foods.Food(traits, {}, arbUnits, 
                    origFile = self.filename)
        return result

    def _isWriteable(self):
        """ False if File instance is read-only, True if read-write. """
        return self._writeable

    writeable = property(fget = _isWriteable, doc = _isWriteable.__doc__)

    def _getGroups(self):
        """ Returns a dictionary of food groups in this cabinet.

        Keys are the food group names; values are lists with
        the keys of foods belonging to that group. 
        """
        import foods
        groupKey = foods.TRAITSBYNAME['group']

        results={}

        for key in self.allTraits.keys():
            groupName = self.allTraits[key][groupKey]
            results.setdefault(groupName, []).append(key)
        return results

    groups = property(fget = _getGroups, doc = _getGroups.__doc__)

    def search(self, patterns, options = 0):
        """ Returns keys of foods matching search terms.

        Search terms are contained in traits, a mapping object.  The
        keys are the traits to search; the values are regexps or
        numbers, as appropriate. Only set the traits you want to
        search on; that is, if you only want to search for names,
        then use searchTraits = {'name':'^Captain Crunch$'}.

        Options are set using global constants IGNORECASE and EXACTMATCH.
        Combine these using a bitwise OR (|). If EXACTMATCH is not
        set, then the function uses regular expressions and the traits
        dict must contain regular expressions. Otherwise, the items
        in the traits dictionary must exactly match the traits in the 
        File instance in order to be returned.
        
        """
        import re
        import foods

        class FoodFilter:
            def reMatcher(self, traitNo, value):
                return self.patterns[traitNo].search(value)

            def exactMatcher(self, traitNo, value):
                return self.patterns[traitNo] == value

            def __init__(self, patterns, options):
                traitdict = foods.TRAITSBYNAME
                self.patterns = dict([(traitdict[patternNo], pattern)
                    for patternNo, pattern
                    in patterns.items()])
                if options & EXACTMATCH:
                    self.matcher = self.exactMatcher
                else:
                    flag = 0
                    if options & IGNORECASE:
                        flag = re.IGNORECASE
                    for patternNo, pattern in self.patterns.copy().items():
                        self.patterns[patternNo] = re.compile(pattern, flag)
                    self.matcher = self.reMatcher

            def __call__(self, traits):
                for patternNo, pattern in self.patterns.items():
                    if not self.matcher(patternNo, traits[patternNo]):
                        return False
                return True

        ff = FoodFilter(patterns, options)
        return [key for key in self.keys() if ff(self.allTraits[key])]


    def _getNuts(self):
        """ Returns mapping for sum of all nuts in this instance. """
        try:
            total = self.values()[0].nuts
        except IndexError:
            total = {}

        for food in self.values()[1:]:
            total = food.nuts + total
        return total

    nuts = property(fget = _getNuts, doc = _getNuts.__doc__)

    def _getGrams(self):
        """ Returns total gram weight of all foods in this instance.
        """

        return sum([food.grams for food in self.values()])

    grams = property(fget = _getGrams, doc = _getGrams.__doc__)

class Shelf(Base):
    """ File instance stored in a shelf; the "native" version.
    """

    def __init__(self, filename, create = False):

        import shelve
        import os
        from bsddb.db import DBAccessError
        from anydbm import error as anydbmError
        from pantry import version
        self.changed = False

        try:
            self.shelf = shelve.open(filename=filename, flag='w')
        except DBAccessError, error:
            if error.args[0] == 13:
                # permission denied
                self.shelf = shelve.open(filename, 'r')
                self.writeable = False
            else:
                raise
        except anydbmError, error:
            if error.args[0] == "need 'c' or 'n' flag to open new db":
                if not create:
                    raise IOError(2, "No such file or directory")
                self.shelf = shelve.open(filename, 'c')
                self.writeable = True
            else:
                raise
        else:
            self.writeable = True

        self.allTraits = self.shelf.get('allTraits', {})
        self.allArbUnits = self.shelf.get('allArbUnits', {})
        shelfVersion = self.shelf.setdefault('VERSION', version.VERSION)
        self.allNuts = self.shelf
        self.filename = os.path.abspath(filename)
        if not version.ACCEPT_VERSION <= shelfVersion <= version.VERSION:
            raise ShelfVersionError

    def save(self):
        if self.changed:
            self.shelf['allTraits'] = self.allTraits
            self.shelf['allArbUnits'] = self.allArbUnits

class Temp(Base):
    """ File instance that starts empty and vanishes when destroyed. """

    def __init__(self):
        self.allTraits = {}
        self.allNuts = {}
        self.allArbUnits = {}
        self.writeable = True
        self.filename = None
        self.origFiles = {}
        self.changed = False

    def addFood(self, food):
        newFood = Base.addFood(self, food)
        self.origFiles[newFood.hash] = food.origFile

    def getFood(self, key, needNuts = True):
        """ Get a food from this file.

        Returns a food with nutrients if needNuts is True;
        otherwise, return a food without nutrients. 
        
        The Temp and Txt getFood is different from the
        others because unlike every other File* instance, Temp
        and Txt stores the actual origFile that each food came
        from."""

        result = Base.getFood(self, key, needNuts)
        result.origFile = self.origFiles[key]
        return result

class Zip(Base):
    """ Holds the contents of a zipped sr database. """

    # Constants from sr:
    # 
    # EOL: end of line character in SR table files.
    # STRINGQUOTE: used to quote string fields in SR.
    # FIELDSEP: SR field separator.
    # ENCODING: hmm...
    # 
    # FIELDS is a dict: 
    #     key: table name
    # 
    #     value: 2-tuple of tuples. value[0] consists of strings, 
    #     with each string being a field name (in the order that they
    #     appear in the table.) value[1] consists of strings, with each 
    #     string being the name of a key field.
    # 
    #     Many key/value pairs in fields are commented out. This way
    #     FileSrZip does not import these tables. In addition,
    #     FileSrZip is unaffected if trailing fields are omitted.
    #     Thus trailing fields are often omitted to save RAM. """

    EOL = '\r\n'
    STRINGQUOTE = '~'
    FIELDSEP = '^'
    ENCODING = 'iso-8859-1'

    # fields. Key: table name. Value: 2-tuple; first tuple
    # are the fields; second tuple is the keys.
    # It does no harm to omit the last fields of a table, so
    # these are omitted where possible to save RAM.
    TABLES = {
        #'food_des':(('ndbNo', 'fdGrpCd', 'longDesc', 'shrtDesc', 'comName', 
        #    'manufacName', 'survey', 'refDesc', 'pctRefuse', 'sciName',
        #    'nFactor', 'proFactor', 'fatFactor', 'choFactor'), ('ndbNo',)),

        'food_des':(('ndbNo', 'fdGrpCd', 'longDesc', 'shrtDesc', 'comName', 
            'manufacName', 'survey', 'refDesc', 'pctRefuse'), ('ndbNo',)),

        'fd_group':(( 'fdGrpCd', 'fdGrpDesc' ), ('fdGrpCd',)),

        # There is a last field in SR, 'cc', but it has not been populated
        # in SR 19. That's what the docs say anyway. It seems there is data
        # there but although it is alphanumeric, it is not surrounded by
        # tildes like all the other strings are. This causes import errors.
        # I have simply omitted the last field, which should not cause any
        # problems. (The last field is 'cc', for 'confidence code.'
        #
        # Subsequently additional fields omitted as well, to save RAM.

        #'nut_data':(( 'ndbNo', 'nutrNo', 'nutrVal', 'numDataPts', 'stdError',
        #    'srcCd', 'derivCd', 'refNdbNo', 'addNutrMark', 'numStudies',
        #    'min', 'max', 'df', 'loweb', 'upeb', 'statcmt'), 
        #    ('ndbNo', 'nutrNo')),

        'nut_data':(( 'ndbNo', 'nutrNo', 'nutrVal'), ('ndbNo', 'nutrNo')),

        #'nutr_def':(( 'nutrNo', 'units', 'tagName', 'nutrDesc', 'numDec',
        #    'srOrder'), ('nutrNo',)),

        #'nutr_def':(( 'nutrNo', 'units', 'tagName', 'nutrDesc'), ('nutrNo',)),

        #'src_cd':(( 'srcCd', 'srcCdDesc'), ('srcCd',)),

        #'deriv_cd':(( 'derivCd', 'derivDesc'), ('derivCd',)),

        #'weight':(( 'ndbNo', 'seq', 'amount', 'measureDesc', 'gmWgt', 'numDataPts',
        #    'stdDev' ), ('ndbNo', 'seq')),

        'weight':(( 'ndbNo', 'seq', 'amount', 'measureDesc', 'gmWgt'), 
            ('ndbNo', 'seq')),

        #'footnote':(( 'ndbNo', 'footntNo', 'footntType', 
        #   'nutrNo', 'footntTxt' ),
        #    ('ndbNo', 'footntNo')),
        #
        #'data_Src':(( 'dataSrcId', 'authors', 'title', 'year', 'journal', 
        #   'volCity', 'issueState', 'startPage', 'endPage' ), ('dataSrcId',)),
        #
        #'datSrcLn':(( 'ndbNo', 'nutrNo', 'dataSrcId' ), ('ndbNo', 'nutrNo'))
        }

    class WrapTable(dict):
        """ Return an object to wrap a database table."""

        def __init__(self, table, fields, keys, eol, fieldSep, stringQuote):
            """ Set up table wrapper.

                Parameters:
                    table: string containing the database table
                    fields: sequence of field names
                    keys: sequence of keys, must be in fieldList
                    eol: string used to indicate end of line
                    fieldSep: string used to separate fields
                    stringQuote: character used to quote strings """

            dict.__init__(self)
            table = table.replace(stringQuote, u'')
            lines = table.rstrip(eol).split(eol)
            columnNumbersAndFieldNames = zip(range(len(fields)), fields)
            for line in lines:
                fileRow=line.split(fieldSep) 
                resultRow = {}
                resultRow = dict([(fieldName, fileRow[columnNumber])
                    for columnNumber, fieldName
                    in columnNumbersAndFieldNames])
                
                resultKey = [ resultRow[key] for key in keys ]
                self[tuple(resultKey)] = resultRow

    def __init__(self, filename):
        import zipfile
        import os
        import errno

        import foods

        self.allTraits = {}
        self.allNuts = {}
        self.allArbUnits = {}
        self.writeable = False
        self.filename = os.path.abspath(filename)
        self.changed = False

        # wrap tables
        z = zipfile.ZipFile(filename, mode = 'r')
        tables={}
        for name, (fields, keys) in self.TABLES.items():
            # convert to unicode; apparently sr is encoded in latin1
            txtRaw = z.read(name.upper() + '.txt')
            txtUni = unicode(txtRaw, self.ENCODING)

            tables[name]=self.WrapTable(txtUni,
                    fields = fields,
                    keys = keys,
                    eol = self.EOL, fieldSep = self.FIELDSEP,
                    stringQuote = self.STRINGQUOTE)
        
        
        allFoodsNdb = {}
        
        # make foods, with traits
        for (ndbNo,), rowFood in tables['food_des'].items():
            group = tables['fd_group'][(rowFood['fdGrpCd'],)]['fdGrpDesc']
            name = rowFood['longDesc']
            traits = { 'name':name, 
                    'group':group,
                    'refDesc':rowFood['refDesc'], 
                    'pctRefuse':rowFood['pctRefuse'],
                    'unit':u'g', 
                    'date':'', 
                    'meal':'', 
                    'comment':'',
                    'order':'',
                    'qty':u'100' }
            allFoodsNdb[ndbNo] = foods.Food(traits, {}, {})

        # add nutrients
        for (ndbNo, nutrNo), rowData in tables['nut_data'].items():
            nutrNo = int(nutrNo)
            if '.' in rowData['nutrVal']:
                nutVal = float(rowData['nutrVal'])
            else:
                nutVal = int(rowData['nutrVal'])
            allFoodsNdb[ndbNo].nuts[nutrNo] = nutVal

        # add arbitrary units
        for (ndbNo, seq), rowUnit in tables['weight'].items():
            if float(rowUnit['amount']) == 1:
                unitDesc = rowUnit['measureDesc']
                if '.' in rowUnit['gmWgt']:
                    unitGrams = float(rowUnit['gmWgt'])
                else:
                    unitGrams = int(rowUnit['gmWgt'])
                allFoodsNdb[ndbNo].arbUnits[unitDesc] = unitGrams

        # add foods
        for food in allFoodsNdb.values():
            self._initFood(food)

class Xml(Base):
    """ Holds a File instance that's stored in an XML file. """

    DTD = """
        <!ELEMENT foods         (food*)>
        <!ELEMENT food          (nutrient+, unit*) >
        <!ELEMENT nutrient      EMPTY>
        <!ELEMENT unit          EMPTY>

        <!ATTLIST food
            name                CDATA #IMPLIED
            group               CDATA #IMPLIED
            pctRefuse           CDATA #IMPLIED
            refDesc             CDATA #IMPLIED 
            date                CDATA #IMPLIED
            meal                CDATA #IMPLIED
            qty                 CDATA #REQUIRED 
            comment             CDATA #IMPLIED
            order               CDATA #IMPLIED
            unit                CDATA #REQUIRED >

        <!ATTLIST nutrient
            name                CDATA #REQUIRED
            units               CDATA #IMPLIED 
            amount              CDATA #REQUIRED >

        <!ATTLIST unit
            grams               CDATA #REQUIRED
            name                CDATA #REQUIRED >"""

    def __init__(self, filename, validate = SMALL, create = False): 
        import os
        import sys
        from xml.sax import make_parser
        from xml.sax.handler import ContentHandler
        import errno

        import pantry.xmlCheck as xmlCheck
        import foods

        self.allTraits = {}
        self.allNuts = {}
        self.allArbUnits = {}
        self.changed = False
        self.filename = os.path.abspath(filename)
        newfile = False

        try:
            self.fh = open(filename, 'r+')
        except IOError, error:
            if error.errno == errno.EACCES: 
                self.fh = open(filename, 'r')
                self.writeable = False
            elif error.errno == errno.ENOENT:
                if not create:
                    raise 
                self.fh = open(filename, 'w')
                self.writeable = True
                newfile = True
            else:
                raise
        else:
            self.writeable = True

        if newfile:
            return

        if xmlCheck.getRootElement(self.fh) != 'foods':
            raise WrongXMLFileError
        self.fh.seek(0)

        # validate XML file if asked to
        try:
            if validate == SMALL:
                xmlCheck.xmlCheck(filename, self.DTD)
            elif validate == ALWAYS:
                xmlCheck.xmlCheck(filename, self.DTD, forceValid = True)
        except xmlCheck.BadXMLError, e:
            raise InvalidXMLError(filename, 'Foods', e)


        class FoodHandler(ContentHandler):
            def __init__(self, cabinet):
                self.cabinet = cabinet

            def startElement(self, name, attrs):
                if name == 'food':
                    self.traits = dict([(attrName, attrValue)
                        for attrName, attrValue
                        in attrs.items()])
                    self.nuts = {}
                    self.arbUnits = {}
                elif name == 'nutrient':
                    nutrVal = attrs['amount']
                    nutName = attrs['name']
                    units = attrs.get('units', None)
                    self.nuts[(nutName, units)] = nutrVal

                elif name == 'unit':
                    self.arbUnits[attrs['name']] = attrs['grams']

            def endElement(self, name):
                if name == 'food':
                    try:
                        newFood = foods.Food(self.traits, self.nuts,
                                self.arbUnits)
                    except foods.Error, e:
                        raise FoodInitError(self.cabinet.filename,
                                self.traits.get('name', 'unknown name'),
                                e)
                    self.cabinet._initFood(newFood)
            
        parser = make_parser()
        parser.setContentHandler(FoodHandler(self))
        parser.parse(self.fh)

    def save(self):
        """ Save the foods to the XML file. """ 
        from pantry.unicodeHelper import puni
        from xml.sax import saxutils
        import foods
        if not self.writeable:
            raise ReadOnlyError

        if not self.changed:
            return

        self.fh.seek(0)
        toFile = puni(self.fh)

        toFile(u'<?xml version="1.0" encoding="%s"?>\n' % toFile.ENCODING)
        toFile(u'<foods>\n')
        for food in self.values():
            toFile(u'<food ')
            for traitName, traitValue in food.traits.items():
                if traitValue:
                    toFile(u'%s = %s ' % (traitName,
                        saxutils.quoteattr(traitValue)))
            toFile(u'>\n')

            for (nutName, nutUnits), nutVal in food.nuts.items():
                toFile(u'\t<nutrient name=%s units=%s amount=%s />\n' % 
                        (saxutils.quoteattr(nutName), 
                            saxutils.quoteattr(nutUnits), 
                            saxutils.quoteattr(str(nutVal))))
                
            for unitName, unitGrams in food.arbUnits.items():
                toFile(u'\t<unit name=%s grams=%s />\n' %
                         (saxutils.quoteattr(unitName), 
                             saxutils.quoteattr(str(unitGrams))))
            toFile(u'</food>\n')

        toFile(u'</foods>\n')
        self.fh.truncate()
        self.fh.flush()

    def __delitem__(self, key):
        if self.flag == 'r':
            raise ReadOnlyError
        Base.__delitem__(self, key)

class Recipes(Base):
    """ Holds a File instance derived from an XML file with recipes.  """

    DTD = """
        <!ELEMENT recipes           (recipe*)>
        <!ELEMENT recipe            (yield, ingredient+, unit*, directions?)>
        <!ELEMENT yield             (#PCDATA)>
        <!ELEMENT ingredient        EMPTY>
        <!ELEMENT unit              EMPTY>
        <!ELEMENT directions        (para+)>
        <!ELEMENT para              (#PCDATA) >

        <!ATTLIST recipes
            foodSource              CDATA #REQUIRED >

        <!ATTLIST recipe
            name                CDATA #IMPLIED
            group               CDATA #IMPLIED
            pctRefuse           CDATA #IMPLIED
            refDesc             CDATA #IMPLIED 
            date                CDATA #IMPLIED
            meal                CDATA #IMPLIED
            qty                 CDATA #REQUIRED 
            comment             CDATA #IMPLIED
            order               CDATA #IMPLIED
            unit                CDATA #REQUIRED >

        <!ATTLIST yield
            grams                   CDATA #REQUIRED 
            servings                CDATA #IMPLIED >

        <!ATTLIST ingredient
            name                    CDATA #REQUIRED
            qty                     CDATA #REQUIRED
            unit                    CDATA #REQUIRED
            comment                 CDATA #IMPLIED >

        <!ATTLIST unit
            grams                   CDATA #REQUIRED
            name                    CDATA #REQUIRED >"""

    def __init__(self, filename, validate=SMALL, searchPath = None):
        import copy
        import os
        import foods
        from pantry import xmlCheck
        from xml.sax import make_parser
        from xml.sax.handler import ContentHandler

        self.allTraits = {}
        self.allNuts = {}
        self.allArbUnits = {}
        self.writeable = False
        self.filename = os.path.abspath(filename)
        self.changed = False

        if xmlCheck.getRootElement(self.filename) != 'recipes':
            raise WrongXMLFileError

        # validate XML file if asked to
        try:
            if validate == SMALL:
                output = xmlCheck.xmlCheck(filename, self.DTD)
            elif validate == ALWAYS:
                output = xmlCheck.xmlCheck(filename, self.DTD, 
                        forceValid = True)
        except xmlCheck.BadXMLError, e:
            raise InvalidXMLError(filename, 'Recipes', e)

        class RecipeHandler(ContentHandler):
            def __init__(self, cabinet, searchPath):
                self.cabinet = cabinet
                self.searchPath = searchPath

            def startElement(self, name, attrs):
                if name == 'recipes':
                    self.foodSourceFileName = attrs['foodSource']
                    try:
                        self.foodSource = getCabinet(self.foodSourceFileName,
                                self.searchPath)
                    except CabinetOpenError, e:
                        raise SourceFileError(self.cabinet.filename,
                                self.foodSourceFileName, e)

                elif name == 'recipe':
                    self.ingredients = Temp()
                    self.arbUnits = {}
                    self.traits = dict([(name, value)
                        for name, value in attrs.items()])
                    self.nuts = {}

                elif name == 'ingredient':
                    sName = attrs['name']
                    qty = attrs['qty']
                    unit = attrs['unit']
                    foodKeys = self.foodSource.search({'name':sName},
                            EXACTMATCH)
                    locator = u'Recipe named %s' % self.traits['name']

                    if len(foodKeys) != 1:
                        raise FoodSearchError(self.cabinet.filename, locator,
                                sName, self.foodSource, foodKeys)

                    foodKey = foodKeys[0]
                    ingredient = self.foodSource.getFood(foodKey, True).copy()
                    try:
                        ingredient.traits['qty'] = qty
                    except foods.Error, e:
                        raise QtyError(self.cabinet.filename, locator, e)

                    try:
                        ingredient.setUnit(unit, EXACTMATCH)
                    except foods.Error, e:
                        raise UnitSearchError(self.cabinet.filename,
                                locator, e)
                    self.ingredients.addFood(ingredient)
                
                elif name == 'yield':
                    if attrs['grams'] == 'guess':
                        self.grams = None
                    else:
                        self.grams = float(attr['grams'])

                    strServings = attrs.get('servings', '')
                    try:
                        self.floatServings = float(strServings)
                    except ValueError:
                        self.floatServings = None

                elif name == 'unit':
                    unitName = attrs['name']
                    self.arbUnits[unitName] = attrs['grams']

            def endElement(self, name):
                if name == 'recipe':
                    if not self.grams:
                        self.grams = self.ingredients.grams
                    if self.floatServings:
                        self.arbUnits['serving'] =\
                                self.grams / self.floatServings
                    try:
                        newFood = foods.Food({'qty':self.grams, 'unit':'g'},
                                self.ingredients.nuts, self.arbUnits)
                    except foods.Error, e:
                        raise FoodInitError(self.cabinet.filename,
                                self.traits.get('name', 'unknown food'), e)

                    try:
                        newFood.setTraits(self.traits)
                    except foods.Error, e:
                        raise FoodInitError(self.cabinet.filename,
                                self.traits.get('name', 'unknown food'), e)
                    self.cabinet._initFood(newFood)

        parser = make_parser()
        parser.setContentHandler(RecipeHandler(self, searchPath))
        parser.parse(self.filename)
        self.changed = False

class Txt(Base):
    """ Holds a File instance derived from a plain text file. """
    COLS = {'source':0, 'name':1, 'qty':2, 'unit':3,
            'date':4, 'meal':5, 'group':6, 'comment':7,
            'options':8 }
    TRAITCOLS = {'name':1, 'qty':2, 'unit':3,
            'date':4, 'meal':5, 'group':6, 'comment':7 }

    def __init__(self, filename, pathList = [], create = False):
        import re
        import os
        import copy
        import sys
        import locale
        import codecs
        import errno

        import foods

        COLS = Txt.COLS
        TRAITCOLS = Txt.TRAITCOLS

        self.allTraits = {}
        self.allNuts = {}
        self.allArbUnits = {}
        self.origFiles = {}
        self.filename = os.path.abspath(filename)
        self.pathList = pathList
        newfile = False
        self.changed = False
        defaultEncoding = locale.getdefaultlocale()[1]

        try:
            self.fh = codecs.open(filename, 'r+', defaultEncoding)
        except IOError, error:
            # permission denied
            if error.errno == errno.EACCES:
                self.fh = codecs.open(filename, 'r', defaultEncoding)
                self.writeable = False
            # file does not exist
            elif error.errno == errno.ENOENT:
                if not create:
                    raise 
                self.fh = codecs.open(filename, 'w', defaultEncoding)
                self.writeable = True
                newfile = True
            # don't handle any other IOError instances
            else:
                raise
        else:
            self.writeable = True
        
        if newfile:
            self.lines = []
            return

        self.lines = self.fh.readlines()

        # split lines into fields, pad lines
        # that are missing fields on the end
        splitLines = []
        lineNo = 0
        for line in self.lines:
            lineNo += 1

            # strip comments
            line = re.sub(r'(?<!\\)#.*$', '', line)
            if re.search(r'^\s*$', line):
                # skip blank lines
                splitLines.append([])
            else:
                splitLineRaw = re.split(r'(?<!\\):', line.rstrip())
                splitLine = []
                for field in splitLineRaw:
                    # raise error if any backslashes not followed by
                    # ':|#|\'
                    if re.search(r'\\(?=:|#|\\)', field):
                        raise BadBackslashError('line number %s' % lineNo)

                    # replace \:, \\, and \#
                    for char in (':', '\\', '#'):
                        newField = field.replace('\\%s' % char, char)
                    splitLine.append(newField)

                # substitute blanks for empty fields
                if len(splitLine) < 2:
                    raise MissingFieldsError('line number %s' % lineNo)
                while len(splitLine) < len(COLS):
                    splitLine.append('')
                splitLines.append(splitLine)

        sources = {}
        lineNo = 0
        for splitLine in splitLines:
            lineNo += 1
            if splitLine == []:
                continue

            sourceName = splitLine[COLS['source']]
            try:
                source = sources[sourceName]
            except KeyError:
                try:
                    sources[sourceName] = getCabinet(sourceName, self.pathList)
                except CabinetOpenError, e:
                    raise SourceFileError(self.filename, sourceName, e)
                source = sources[sourceName]

            traits = dict([(colName, splitLine[colNo])
                for colName, colNo in TRAITCOLS.items() 
                if splitLine[colNo] != ''])
            sName = traits.pop('name')
            traits['order'] = '%04u' % lineNo

            options = 0
            for letter in splitLine[COLS['options']]:
                if letter == 'i':
                    options = options | IGNORECASE
                elif letter == 'x':
                    options = options | EXACTMATCH
                else:
                    raise UnrecognizedOptionError('line number %s' % lineNo,
                            letter)

            foodKeys = source.search({'name':sName}, options)

            if len(foodKeys) != 1:
                raise FoodSearchError(self.filename, 
                        'line number %s' % lineNo, sName,
                        source, foodKeys)

            foodKey = foodKeys[0]
            food = source.getFood(foodKey, True).copy()
            for traitName, traitValue in traits.items():
                if traitName == 'unit':
                    try:
                        food.setUnit(traits['unit'], options)
                    except foods.SetUnitError, e:
                        raise UnitSearchError(self.filename,
                                'line number %s' % lineNo, e)
                else:
                    try:
                        food.traits[traitName] = traitValue
                    except foods.TraitValueError, e:
                        raise QtyError(self.filename,
                                'line number %s' % lineNo, e)

            self._initFood(food)

    def __delitem__(self, key):
        lineNoToDelStr = self[key].getTraits()['order']
        lineNoToDel = int(lineNoToDelStr.lstrip('0'))

        # decrement line number of all subsequent foods
        for food in self.values():
            afterLineNoStr = food.getTraits()['order']
            afterLineNo = int(afterLineNoStr.lstrip('0'))
            if afterLineNo > lineNoToDel:
                afterLineNo -= 1
                food.changeTraits({'order': 
                    u'%04u' % afterLineNo})
        
        del self.lines[lineNoToDel - 1]
        del self.origFiles[key]
        Base.__delitem__(self, key)

    def save(self):
        if not self.changed:
            return
        self.fh.seek(0)
        self.fh.writelines(self.lines)

    def getFood(self, key, needNuts = True):
        """ Get a food from this file.

        Returns a food with nutrients if needNuts is True;
        otherwise, return a food without nutrients. 
        
        The Temp and Txt getFood is different from the
        others because unlike every other File* instance, Temp
        and Txt stores the actual origFile that each food came
        from."""

        result = Base.getFood(self, key, needNuts)
        return result

    def _initFood(self, food):
        newHash = food.hash
        (self.allTraits[newHash], self.allNuts[newHash],
            self.allArbUnits[newHash], self.origFiles[newHash]) = food.data

    def addFood(self, food, incrementOrder = False):
        """ incrementOrder has no effect with Txt cabinets. """
        import os

        if not self.writeable:
            raise ReadOnlyError

        food = food.copy()
        traits = food.traits
        traits['order'] = (u'%04u' % (len(self.lines) + 1))

        pathList = self.pathList
        if not pathList:
            pathList = []
        origFile = food.origFile
        if os.path.dirname(origFile) in pathList:
            origFile = os.path.basename(origFile)

        line = [origFile,
            traits['name'],
            traits['qty'],
            traits['unit'],
            traits['date'],
            traits['meal'],
            traits['group'],
            traits['comment'],
            'x\n' ]
        for char in (':', '#', '\\'):
            newLine = [ field.replace(char, '\\%s' % char) for field in line ]
        self.lines.append(':'.join(newLine))
        self._initFood(food) 
        self.changed = True

def getCabinet(filename, pathList = None, validate = SMALL, create = False):
    """ Get a File object appropriate for filename. 

    getCabinet behaves like shells do when searching for filename;
    that is, if filename contains os.sep, then getCabinet will not
    search the path at all but will instead treat filename as an
    absolute path. Otherwise, getCabinet searches the directories in
    pathList for filename.  pathList is a list of directories to
    search.  Indicate the current directory with the empty string
    (''). If pathList is not set, getCabinet uses the current
    directory only. """


    import os
    import sys
    from pantry import xmlCheck
    import errno

    if not pathList:
        pathList = [ os.getcwd() ]

    if os.sep in filename:
        possFiles = [ filename ]
    else:
        possFiles = [ os.path.join(dir, filename) for dir in pathList ]

    ext = os.path.splitext(filename)[1]
    for possFile in possFiles:
        if ext.lower() == '.zip':
            try:
                result = Zip(possFile)
            except IOError, e:
                if e.errno == errno.ENOENT:
                    continue
                else:
                    raise
            return result
        elif ext.lower() == '.txt':
            try:
                result = Txt(possFile, pathList = pathList)
            except IOError, e:
                if e.errno == errno.ENOENT:
                    continue
                else:
                    raise
            return result
        elif ext.lower() == '.xml':
            try:
                rootElement = xmlCheck.getRootElement(possFile)
            except xmlCheck.NoRootElementError:
                raise UnknownXMLError(possFile)
            except xmlCheck.FileNotFoundError:
                continue 
            if rootElement == 'foods':
                result = Xml(possFile, validate)
            elif rootElement == 'recipes':
                result = Recipes(possFile, validate, pathList)
            else:
                raise WrongXMLError(possFile)
            return result

        else:
            try:
                result = Shelf(possFile)
            except IOError, e:
                if e.errno == errno.ENOENT:
                    continue
                else:
                    raise
            return result
    
    # if we've gotten this far, file not found
    if create:
        if ext.lower() == '.xml':
            return Xml(filename, NEVER, True)
        elif ext.lower() == '.txt':
            return Txt(filename, pathList, True)
        else:
            return Shelf(filename, True)
    else:
        raise FileNotFoundError(filename)

class Error(Exception):
    pass

class CabinetOpenError(Error):
    """ Root class for all errors that can occur when opening a cabinet.

    Subclasses must define a msg attribute--a unicode indicating
    the nature of the error. In addition, subclasses must define a 
    filename attribute, which is a unicode indicating the filename
    in which the error occurred. """
    
    def __init__(self, filename):
        self.msg = []

class SourceFileError(CabinetOpenError):
    """ A problem occurred in opening the source file for Recipes or Txt. """
    def __init__(self, filename, sourceFile, lowerException):
        CabinetOpenError.__init__(self, filename)
        self.msg.append(u'Could not open source file %s' % sourceFile)
        self.msg.extend(lowerException.msg)

class XMLError(CabinetOpenError):
    """ XML is invalid, the wrong type, not well formed, etc.

    Subclasses have specific errors. XMLError should be used as a base
    class only if the problem is with the XML itself--that is, the XML
    is not valid, it is not XML at all, or it is not a Pantry XML file.
    If the XML is perfectly valid but there are other problems with it 
    (e.g. the foods in it are wrong) do not subclass this. """
    def __init__(self, filename):
        CabinetOpenError.__init__(self, filename)
        self.msg.append('Could not read XML:')

class UnknownXMLError(XMLError):
    """ Root element of XML cannot be determined. """

    def __init__(self, filename):
        XMLError.__init__(self, filename)
        self.msg.append(u'Root element cannot be determined.')

class InvalidXMLError(XMLError):
    """ XML file is invalid. """

    def __init__(self, filename, xmlType, errorInst):
        """ xmlType: unicode, either Foods or Recipes as appropriate. """
        XMLError.__init__(self, filename)
        self.msg.append(u'File does not validate as a %s file.' % xmlType)
        self.msg.extend(errorInst.msg)

class WrongXMLError(XMLError):
    """ XML is not Pantry. """

    def __init__(self, filename):
        XMLError.__init__(self, filename)
        self.msg.append('XML file is not a Pantry file.')

class InvalidDataError(CabinetOpenError):
    def __init__(self, locator):
        self.msg = [u'Invalid data in file at location: %s' % locator]

class FoodSearchError(InvalidDataError):
    """ Search for foods for ingredient or Txt returned != 1 match. """
    def __init__(self, filename, locator, pattern, source, keys):
        InvalidDataError.__init__(self, locator)
        self.msg.append('Search pattern %s did not match exactly one food.'
                % pattern)
        self.msg.append('Matches:')
        if len(keys) > 1:
            matchingNames = [source.getFood(key, False).traits['name']
                    for key in keys]
            self.msg.append(u'\n'.join(matchingNames))
        else:
            self.msg.append(u'<none>')

class QtyError(InvalidDataError):
    """ Qty for an ingredient or FoodsTxt food would not convert to float. """
    def __init__(self, filename, locator, foodsError):
        InvalidDataError.__init__(self, locator)
        self.msg.extend(foodsError.msg)

class UnitSearchError(InvalidDataError):
    """ Search for unit for ingredient or for Foods Txt has != 1 match. """
    def __init__(self, filename, locator, foodsError):
        InvalidDataError.__init__(self, locator)
        self.msg.extend(foodsError.msg)

class FoodInitError(InvalidDataError):
    """ Could not initialize food in XML or Recipes file. """
    def __init__(self, filename, locator, foodsError):
        InvalidDataError.__init__(self, locator)
        self.msg.extend(foodsError.msg)

class TxtError(InvalidDataError):
    """ Problems parsing Txt files. """
    def __init__(self, locator):
        InvalidDataError.__init__(self, locator)

class BadBackslashError(TxtError):
    def __init__(self, locator):
        TxtError.__init__(self, locator)
        self.msg.append('Backslash not followed by "#", ":", or "\\"')

class MissingFieldsError(TxtError):
    def __init__(self, locator):
        TxtError.__init__(self, locator)
        self.msg.append('Each line must have at least two fields.')

class UnrecognizedOptionError(TxtError):
    def __init__(self, locator, option):
        TxtError.__init__(self, locator)
        self.msg.append('Option %s not recognized.' % option)

class FileNotFoundError(CabinetOpenError):
    def __init__(self, filename):
        CabinetOpenError.__init__(self, filename)
        self.msg.append(u'File not found.')

class ReadOnlyError(Error):
    pass
