# Pantry, command-line nutrient analysis. Version 11
# Copyright 2007 Omari Norman. 
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE. 
# 
# Pantry's "built-in" XML validator is xmlproc.
# xmlproc is Copyright 1998-2000 by Lars Marius Garshol, Oslo, Norway. 
# 
# All Rights Reserved 
# 
# Permission to use, copy, modify, and distribute this software
# and its documentation for any purpose and without fee is hereby
# granted, provided that the above copyright notice appear in all
# copies and that modified copies are clearly marked as such. 
# 
# LARS MARIUS GARSHOL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
# SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS, IN NO EVENT SHALL LARS MARIUS GARSHOL BE LIABLE FOR
# ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
# IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
# ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
# THIS SOFTWARE. 
""" 
Food, Foods and associated classes - the core of Pantry

This module contains the classes that comprise the core of Pantry.
The Food object holds information about a single food--its name,
quantity, nutrients, and so forth. It also has 'date' and 'meal'
attributes. Ordinarily the user will keep some sort of 'master' list
of foods, and both 'date' and 'meal' will be empty strings.

"""

from __future__ import division
import UserDict

# Constants 
IGNORECASE = 1
EXACTMATCH = 2

TRAITNAMES = ('name', 'group', 'refDesc', 'pctRefuse',
        'unit', 'date', 'meal', 'qty', 'comment', 'order')
FLOATTRAITS = ('qty', 'pctRefuse')
TRAITSBYNO = dict(zip(range(len(TRAITNAMES)), TRAITNAMES))
TRAITSBYNAME = dict(zip(TRAITNAMES, range(len(TRAITNAMES))))

NUTSBYNO = {
    203: ('Protein', 'g', 50),
    204: ('Total Fat', 'g', 65),
    205: ('Total Carbohydrate', 'g', 300),
    207: ('Ash', 'g', None),
    208: ('Calories', 'kcal', None),
    209: ('Starch', 'g', None),
    210: ('Sucrose', 'g', None),
    211: ('Glucose (dextrose)', 'g', None),
    212: ('Fructose', 'g', None),
    213: ('Lactose', 'g', None),
    214: ('Maltose', 'g', None),
    221: ('Alcohol, ethyl', 'g', None),
    255: ('Water', 'g', None),
    257: ('Adjusted Protein', 'g', None),
    262: ('Caffeine', 'mg', None),
    263: ('Theobromine', 'mg', None),
    268: ('Energy', 'kj', None),
    269: ('Sugars', 'g', None),
    287: ('Galactose', 'g', None),
    291: ('Dietary Fiber', 'g', 25),
    301: ('Calcium', 'mg', 1000),
    303: ('Iron', 'mg', 18),
    304: ('Magnesium', 'mg', 400),
    305: ('Phosphorus', 'mg', 1000),
    306: ('Potassium', 'mg', 3500),
    307: ('Sodium', 'mg', 2400),
    309: ('Zinc', 'mg', 15),
    312: ('Copper', 'mg', 2),
    313: ('Fluoride', 'mcg', None),
    315: ('Manganese', 'mg', None),
    317: ('Selenium', 'mcg', None),
    318: ('Vitamin A', 'IU', 5000),
    319: ('Retinol', 'mcg', None),
    320: ('Vitamin A, RAE', 'mcg_RAE', None),
    321: ('Carotene, beta', 'mcg', None),
    322: ('Carotene, alpha', 'mcg', None),
    323: ('Vitamin E', 'mg', None),
    324: ('Vitamin D', 'IU', 400),
    334: ('Cryptoxanthin, beta', 'mcg', None),
    337: ('Lycopene', 'mcg', None),
    338: ('Lutein + zeaxanthin', 'mcg', None),
    341: ('Tocopherol, beta', 'mg', None),
    342: ('Tocopherol, gamma', 'mg', None),
    343: ('Tocopherol, delta', 'mg', None),
    401: ('Vitamin C', 'mg', 60),
    404: ('Thiamin', 'mg', 1.5),
    405: ('Riboflavin', 'mg', 1.7),
    406: ('Niacin', 'mg', 20),
    410: ('Pantothenic acid', 'mg', 10),
    415: ('Vitamin B-6', 'mg', 2),
    417: ('Folate, total', 'mcg', None),
    418: ('Vitamin B-12', 'mcg', 6),
    421: ('Choline, total', 'mg', None),
    430: ('Vitamin K (phylloquinone)', 'mcg', None),
    431: ('Folic acid', 'mcg', 400),
    432: ('Folate, food', 'mcg', None),
    435: ('Folate, DFE', 'mcg_DFE', None),
    454: ('Betaine', 'mg', None),
    501: ('Tryptophan', 'g', None),
    502: ('Threonine', 'g', None),
    503: ('Isoleucine', 'g', None),
    504: ('Leucine', 'g', None),
    505: ('Lysine', 'g', None),
    506: ('Methionine', 'g', None),
    507: ('Cystine', 'g', None),
    508: ('Phenylalanine', 'g', None),
    509: ('Tyrosine', 'g', None),
    510: ('Valine', 'g', None),
    511: ('Arginine', 'g', None),
    512: ('Histidine', 'g', None),
    513: ('Alanine', 'g', None),
    514: ('Aspartic acid', 'g', None),
    515: ('Glutamic acid', 'g', None),
    516: ('Glycine', 'g', None),
    517: ('Proline', 'g', None),
    518: ('Serine', 'g', None),
    521: ('Hydroxyproline', 'g', None),
    573: ('Vitamin E, added', 'mg', None),
    578: ('Vitamin B-12, added', 'mcg', None),
    601: ('Cholesterol', 'mg', 300),
    605: ('Trans Fat', 'g', None),
    606: ('Saturated Fat', 'g', 20),
    607: ('4:0', 'g', None),
    608: ('6:0', 'g', None),
    609: ('8:0', 'g', None),
    610: ('10:0', 'g', None),
    611: ('12:0', 'g', None),
    612: ('14:0', 'g', None),
    613: ('16:0', 'g', None),
    614: ('18:0', 'g', None),
    615: ('20:0', 'g', None),
    617: ('18:1 undifferentiated', 'g', None),
    618: ('18:2 undifferentiated', 'g', None),
    619: ('18:3 undifferentiated', 'g', None),
    620: ('20:4 undifferentiated', 'g', None),
    621: ('22:6 n-3', 'g', None),
    624: ('22:0', 'g', None),
    625: ('14:1', 'g', None),
    626: ('16:1 undifferentiated', 'g', None),
    627: ('18:4', 'g', None),
    628: ('20:1', 'g', None),
    629: ('20:5 n-3', 'g', None),
    630: ('22:1 undifferentiated', 'g', None),
    631: ('22:5 n-3', 'g', None),
    636: ('Phytosterols', 'mg', None),
    638: ('Stigmasterol', 'mg', None),
    639: ('Campesterol', 'mg', None),
    641: ('Beta-sitosterol', 'mg', None),
    645: ('Monounsaturated Fat', 'g', None),
    646: ('Polyunsaturated Fat', 'g', None),
    652: ('15:0', 'g', None),
    653: ('17:0', 'g', None),
    654: ('24:0', 'g', None),
    662: ('16:1 t', 'g', None),
    663: ('18:1 t', 'g', None),
    664: ('22:1 t', 'g', None),
    665: ('18:2 t not further defined', 'g', None),
    666: ('18:2 i', 'g', None),
    669: ('18:2 t,t', 'g', None),
    670: ('18:2 CLAs', 'g', None),
    671: ('24:1 c', 'g', None),
    672: ('20:2 n-6 c,c', 'g', None),
    673: ('16:1 c', 'g', None),
    674: ('18:1 c', 'g', None),
    675: ('18:2 n-6 c,c', 'g', None),
    676: ('22:1 c', 'g', None),
    685: ('18:3 n-6 c,c,c', 'g', None),
    687: ('17:1', 'g', None),
    689: ('20:3 undifferentiated', 'g', None),
    693: ('Fatty acids, total trans-monoenoic', 'g', None),
    695: ('Fatty acids, total trans-polyenoic', 'g', None),
    696: ('13:0', 'g', None),
    697: ('15:1', 'g', None),
    851: ('18:3 n-3 c,c,c', 'g', None),
    852: ('20:3 n-3', 'g', None),
    853: ('20:3 n-6', 'g', None),
    855: ('20:4 n-6', 'g', None),
    856: ('18:3i', 'g', None),
    857: ('21:5', 'g', None),
    858: ('22:4', 'g', None) }

NUTSBYNAME = dict([(name, (number, units, dv))
    for number, (name, units, dv) in NUTSBYNO.items() ])

ABSOLUTES = { u'g':1.0, u'oz':28.35, u'lb':453.59 }

class _Nuts(UserDict.DictMixin):
    """ Contains several nutrients.
        
    A dictionary; the keys are nutrientNo,
    and the values are the amount of the nutrient (float). """

    def __init__(self, parent, nuts = None):
        """ Set up new Nuts instance.

        To make empty Nuts instance, leave nuts = None.
        You may also use a mapping instance or a sequence of
        2-tuples, as you can with ordinary dictionaries. """

        self.nutdict = {}
        self.parent = parent
        try:
            nuts = nuts.items()
        except AttributeError:
            pass

        self.setitems(nuts)

    def __add__(self, other):
        """ Add two Nuts instances together. """

        commonNuts = set(self).intersection(set(other))
        results = dict([(key, self[key] + other[key])
            for key in commonNuts])
        results.update([(key, self[key]) for key in self
            if key not in commonNuts])
        results.update([(key, other[key]) for key in other
            if key not in commonNuts])

        return results

    def __getitem__(self, key):
        """ Given key, returns nutrient value. 
        
        Key is a sequence. First value is the name of the nutrient;
        second value is the units. Units is optional (currently, including
        them has no effect. """

        nutNo = NUTSBYNAME[key[0]][0]
        return self.nutdict[nutNo] * self.parent.grams / 100 

    def setitems(self, items):
        """ Adds nutrients to this instance.

        items: mapping with nutrients to add. Keys
        may be integers of the nutrient numbers. Or,
        keys may be 2-tuples, with the first value being the name
        of the nutrient and the second value being the units. The
        second value is optional. If it is '%', value is converted according 
        to %DV. If it is anything else non-zero-length, it is 
        checked to make sure it matches the nutrient name. If it is
        None, if it is zero-length, or if it does not exist, no
        checking of units is done.

        Thus, the following are permissible as keys:
            203
            ('Protein',)
            ('Protein', '%')
            ('Protein', 'g')

        The following are NOT permissible as keys:
            (203,)
            'Protein'

        The values are what you want to set the nutrient equal to.
        The value should correspond to the amount of the nutrient
        for the total amount of the food; setitems internally
        converts to nutrients per 100g. 

        If key is an integer, value must be an integer or a float.
        If key is a tuple, value must be a string or Unicode. """

        from pantry import unicodeHelper

        grams = self.parent.grams
        for key, value in items:
            if key not in NUTSBYNO:
                try:
                    value = unicodeHelper.convertUni(value)
                except unicodeHelper.UnconvertableError:
                    raise NutValueError(key, value)
            
                try:
                    nutNo = NUTSBYNAME[key[0]][0]
                except KeyError:
                    raise NutIDError(key, value)

                try:
                    nutUnits = key[1]
                except IndexError:
                    nutUnits = None

                if nutUnits == '%':
                    dvValue = NUTSBYNO[nutNo][2]
                    if dvValue == None:
                        raise NoPercentageError(key, value)
                    value = value / 100 * dvValue
                elif nutUnits and nutUnits != NUTSBYNO[nutNo][1]:
                    raise NutUnitMismatchError(key, value, NUTSBYNO[nutNo][1])

                value = value / grams * 100
                key = nutNo
            self.nutdict[key] = value

    def __setitem__(self, key, value):
        self.setitems(((key, value),))

    def __delitem__(self, key):
        """ Key is a sequence, (nutName, nutUnits).

        Currently the units has no effect though. """

        nutNo = NUTSBYNAME[key[0]][0]
        del self.nutdict[nutNo]
    
    def keys(self):
        return [ (NUTSBYNO[nutNo][0], NUTSBYNO[nutNo][1]) 
                for nutNo in self.nutdict ]

class _Traits(UserDict.DictMixin):
    """Contains a food's traits.

    From the outside this looks like a dictionary, with the keys
    being the trait names, and the values being the trait values. 
    On the inside, the trait names are stored as integers to save
    space. """

    def __init__(self, parent, initial = None):
        """ Set up new Traits instance. """

        self.parent = parent

        # initialize traits 
        self.traitdict = dict([(traitNo, u'') for traitNo
            in TRAITSBYNO])
        del self.traitdict[TRAITSBYNAME['qty']]
        del self.traitdict[TRAITSBYNAME['unit']]

        try:
            initial = initial.items()
        except AttributeError:
            pass

        self.setitems(initial)

        try:
            self['qty']
        except KeyError:
            raise NoQtyError()
        try:
            self['unit']
        except KeyError:
            raise NoUnitError()

    def setitems(self, items):
        """ Set traits in items.

        Items is a mapping. Keys may be integers corresponding
        to those in TRAITSBYNO, or they may be strings
        corresponding to those in TRAITSBYNAME. Values
        are always Unicodes, even for qty and pctRefuse. """

        from pantry import unicodeHelper

        for key, value in items:
            if key not in TRAITSBYNO:
                if key in FLOATTRAITS:
                    try:
                        unicodeHelper.convertUni(value)
                    except unicodeHelper.UnconvertableError:
                        raise TraitValueError(key, value)
                if key == 'unit' and value not in self.parent.units:
                    raise NoAvailUnitError(key, value)
                key = TRAITSBYNAME[key]
            self.traitdict[key] = value

            # update parent grams if necessary
            if key == TRAITSBYNAME['unit'] or key == TRAITSBYNAME['qty']:
                try:
                    qtyStr = self['qty']
                except KeyError:
                    pass
                else:
                    try:
                        qtyNo = unicodeHelper.convertUni(qtyStr)
                    except unicodeHelper.UnconvertableError:
                        raise TraitValueError(key, value)

                    try:
                        gramsPerUnit = self.parent.units[self['unit']]
                    except KeyError:
                        pass
                    else:
                        self.parent._grams = (qtyNo * gramsPerUnit)

    def __setitem__(self, key, value):
        self.setitems(((key, value),))

    def __getitem__(self, key):

        traitNo = TRAITSBYNAME[key]
        return self.traitdict[traitNo]

    def __delitem__(self, key):
        try:
            traitNo = TRAITSBYNAME[key]
        except KeyError:
            traitNo = key
        del self.traitdict[traitNo]

    def keys(self):
        return [ traitName for traitName, traitNo
                in TRAITSBYNAME.items()
                if traitNo in self.traitdict ]

class _ArbUnits(UserDict.DictMixin):
    """ Contains a food's arbitrary units. """
    def __init__(self, parent, initial = None):
        self.arbUnitDict = {}
        self.parent = parent
        try:
            initial = initial.items()
        except AttributeError:
            pass

        self.setitems(initial)

    def setitems(self, items):

        from pantry import unicodeHelper

        for key, value in items:

            # the dreaded isinstance
            if isinstance(value, basestring):
                try:
                    value = unicodeHelper.convertUni(value)
                except unicodeHelper.UnconvertableError:
                    raise ArbUnitValueError(key, value)
            self.arbUnitDict[key] = value

    def __setitem__(self, key, value):
        self.setitems(((key, value),))

    def __getitem__(self, key):
        return self.arbUnitDict[key]

    def __delitem__(self, key):
        del self.arbUnitDict[key]

    def keys(self):
        return self.arbUnitDict.keys()

class Food(object):
    """ Defines a food. """

    def __init__(self, traits, nuts, arbUnits, origFile = None):
        """ Set up a new food.

        Arguments:
            origFile: string, file this food originally came from.
        """

        self._arbUnits = _ArbUnits(self, arbUnits)
        self._traits = _Traits(self, traits)
        self._nuts = _Nuts(self, nuts = nuts)
        self._origFile = origFile

    def __str__(self):
        import pprint
        rep = 'Traits:\n%s\n' % pprint.pformat(self.traits)
        rep = rep + 'Gram weight: %s\n' % self.grams
        rep = rep + 'nuts:\n%s\n' % pprint.pformat(self.nuts)
        rep = rep + 'Units:\n%s\n' % pprint.pformat(self.units)
        return rep

    def _getData(self):
        """ Tuple with data that may be used to reconstruct this food.

        index 0: trait items
        index 1: nut items
        index 2: arbUnit items
        index 3: origFile (unicode, or None)
        """

        return (self._traits.traitdict.copy(), 
                self._nuts.nutdict.copy(),
                self._arbUnits.arbUnitDict.copy(), self._origFile)

    data = property(fget = _getData, doc = _getData.__doc__)

    def copy(self):
        """ Return a deep copy of this food. """
        data = self.data
        return Food(data[0], data[1], data[2], data[3])

    def _getGrams(self):
        return self._grams

    grams = property(fget = _getGrams,
            doc = "Weight of this food, considering quantity and units.")

    def _getNuts(self):
        return self._nuts

    nuts = property(fget = _getNuts,
            doc = """Mapping of nutrients in this food.

            Keys are (nutName, nutUnits); values are nutrient amounts. """)

    def _getArbUnits(self):
        return self._arbUnits

    arbUnits = property(fget = _getArbUnits,
            doc = """ Mapping of arbitrary units in this food.

            Keys are unit names; values are amounts. """)

    def _getOrigFile(self):
        return self._origFile

    def _setOrigFile(self, filename):
        self._origFile = filename

    def _delOrigFile(self):
        del self._origFile

    origFile = property(fget = _getOrigFile,
            fset = _setOrigFile,
            fdel = _delOrigFile,
            doc = """The file this food originally came from.""")

    def _getUnits(self):
        return dict(self._arbUnits.items() + ABSOLUTES.items())

    units = property(fget = _getUnits,
        doc = """ Gets units available for this food.

            Absolute units "mask" arbitrary units--that is,
            if there is an absolute unit and an arbitrary unit
            with the same description, getUnits returns only
            the absolute unit.  """)

    def _getTraits(self):
        return self._traits

    traits = property(fget = _getTraits,
        doc = """ Traits of this food.

        Traits are set using unicode, even for
        floating-point traits. """)

    def _getQtyMinusRefuse(self):
        """ Gets the quantity of this food, reduced by its percent refuse. """

        pctRefuse = float(self.traits['pctRefuse'])
        oldQty = float(self.traits['qty'])
        try:
            newQty = oldQty - pctRefuse / 100 * oldQty
        except TypeError:
            # pctRefuse is None
            newQty = None
        return str(newQty)

    qtyMinusRefuse = property(fget = _getQtyMinusRefuse,
            doc = """ Quantity of food, reduced by percent refuse.

            If percent refuse is None, this is None as well. """)

    def _getHash(self):
        """ Hash of this food.

        The hash is only of the food's traits; nuts and arbUnits are
        not considered when calculating the hash. 
        
        Unicode strings are first encoded to UTF-8, as one cannot
        hash Unicode objects directly; they must be converted
        to bits."""        
        
        # hashlib new in python-2.5
        try:
            import hashlib
            hasher = hashlib.sha1
        except ImportError:
            import sha 
            hasher = sha.sha

        # why not just use self.traits.values()? Because
        # who knows what order they'd be in, and that
        # would affect the hash. This way, the traits
        # are always in a specified order
        fingerPrint = ''
        for traitName in TRAITNAMES:
            fingerPrint += self._traits[traitName].encode('utf-8')
        return hasher(fingerPrint).hexdigest()

    hash = property(fget = _getHash,
            doc = _getHash.__doc__)

    def getQtyByNut(self, sNutName, nutAmt, options = 0):
        """ Return qty of food needed for amount of nutName to equal nutQty.

        options are set using global constants IGNORECASE and EXACTMATCH.
        getQtyByNut looks up nutName using these options. 
        
        nutAmt is a Unicode. """

        from pantry import unicodeHelper

        nutNames = [ key[0] for key in self.nuts.keys() ]
        matches = _listSearch(nutNames, sNutName, options)
        if len(matches) != 1:
            raise NutrientSearchError(self, sNutName, matches, nutAmt)
        nutName = matches[0]
        nutValue = self.nuts[[nutName]]
        try:
            nutAmt = unicodeHelper.convertUni(nutAmt)
        except unicodeHelper.UnconvertableError:
            raise BadAmtError(self, sNutName, matches, nutAmt)

        foodQty = ((unicodeHelper.convertUni(self.traits['qty']) * nutAmt) 
                / nutValue)
        return str(round(foodQty,2))

    def setTraits(self, newTraits):
        """ Set traits en masse. """
        try:
            newTraits = newTraits.items()
        except AttributeError:
            pass

        for key, value in newTraits:
            self.traits[key] = value

    def setUnit(self, pattern, options):
        """ Set unit of this food to the unit matching pattern. """
        matches = _listSearch(self.units, pattern, options)
        if len(matches) != 1:
            raise SetUnitError(self, pattern, matches) 
        self.traits['unit'] = matches[0]

class CmpFoods(object):
    """ Compare two foods for sorting purposes. """

    def __init__(self, traits, sortOrders):
        self.traits = traits
        self.sortOrders = sortOrders

    def __call__(self, x, y):

        for trait in self.traits:
            invert = False
            if trait.startswith('-'):
                invert = True
                trait = trait.lstrip('-')

            xTrait = x.traits[trait]
            yTrait = y.traits[trait]
            if trait in self.sortOrders:
                sortOrder = self.sortOrders[trait]
                if xTrait in sortOrder and yTrait in sortOrder:
                    result = cmp(sortOrder.index(xTrait),
                            sortOrder.index(yTrait))
                elif xTrait not in sortOrder and yTrait not in sortOrder:
                    result = cmp(xTrait, yTrait)
                elif xTrait in sortOrder and yTrait not in sortOrder:
                    result = 1
                elif xTrait not in sortOrder and yTrait in sortOrder:
                    result = -1
            else:
                result = cmp(xTrait, yTrait)
            
            if invert:
                result = -result
            if result != 0:
                return result
        return 0

def _listSearch(sList, sStr, options = 0):
    """ Returns list of items from sList that match sStr.
    
    Options are set using the global constants IGNORECASE
    and EXACTMATCH."""

    if options & EXACTMATCH:
        return [ item for item in sList if item == sStr ]
    else:
        import re
        flag = 0
        if options & IGNORECASE:
            flag = re.IGNORECASE
        sRE = re.compile(sStr, flag)

        return [ item for item in sList if sRE.search(item) ]

class Error(Exception):
    """ Base class for exceptions in this module."""
    pass

class SetUnitError(Error):
    """ Could not set unit to unit requested. """
    def __init__(self, food, pattern, matches):
        self.msg  = [u'Failed to set units using pattern %s' % pattern]
        self.msg.append('Matches:')
        if len(matches) > 0:
            self.msg.append('\n'.join(matches))
        else:
            self.msg.append('<none>')

class QtyByNutError(Error):
    """ getQtyByNut errors. """
    def __init__(self, food, pattern, matches, nutAmt):
        self.msg = [u'Failed to find qty of food per amount nutrient.']

class NutrientSearchError(QtyByNutError):
    """ getQtyByNut search returned != 1 match. """
    def __init__(self, food, pattern, matches, nutAmt):
        QtyByNutError.__init__(self, food, pattern, matches, nutAmt)
        self.msg.append(u'Pattern %s did not yield exactly one match.'
                % pattern)
        self.msg.append(u'Matches:')
        if len(matches) > 1:
            self.msg.append(u'\n'.join(matches))
        else:
            self.msg.append(u'<none>')

class BadAmtError(QtyByNutError):
    """ getQtyByNut failed because amount would not convert. """
    def __init__(self, food, pattern, matches, nutAmt):
        QtyByNutError.__init__(self, food, pattern, matches, nutAmt)
        self.msg.append('Value %s is not a number.' % nutAmt)

class SetNutError(Error):
    """ Errors that occur in Nuts.__setitem__. """
    def __init__(self, key, value):
        self.key = key
        self.value = value
        self.msg = [u'Failed to set nutrient %s' % key[0]]

class NutIDError(SetNutError):
    """ Nutrient ID not found.

    Raised if nutrient number is not found, or if nutrient name is
    not found if nutrient is being set by name. Do not raise if nutrient
    number or name is found, but units do not match. """
    def __init__(self, key, value):
        SetNutError.__init__(self, key, value)
        self.msg.append('Nutrient name is invalid.')

class NutValueError(SetNutError):
    """ Nutrient value is erroneous.

    e.g. nutrient value will not convert to a float. """

    def __init__(self, key, value):
        SetNutError.__init__(self, key, value)
        self.msg.append('Nutrient value %s is not a number.' % value)

class NoPercentageError(SetNutError):
    """ Units were set to percentage, but there's no %DV for this nut. """
    
    def __init__(self, key, value):
        SetNutError.__init__(self, key, value)
        self.msg.append('No percent daily value available.')

class NutUnitMismatchError(SetNutError):
    """ Units given do not match the units for this nutrient. """
    
    def __init__(self, key, value, validUnit):
        SetNutError.__init__(self, key, value)
        self.msg.append('Given unit: %s Valid unit: %s' % (key[1], validUnit))

class SetTraitError(Error):
    """ Errors that occur in Traits.__setitem__. """
    def __init__(self, key, value):
        self.key = key
        self.value = value
        self.msg = [u'Failed to set trait %s' % key]

class NoAvailUnitError(SetTraitError):
    """ No available unit for given unit. """
    def __init__(self, key, value):
        SetTraitError.__init__(self, key, value)
        self.msg.append(u'Unit not available: %s' % value)

class TraitValueError(SetTraitError):
    """ Trait value will not convert to float. """
    
    def __init__(self, key, value):
        SetTraitError.__init__(self, key, value)
        self.msg.append(u'Value %s is not a number' % value)

class TraitInitError(Error):
    """ Problems initializing a food's traits. """
    def __init__(self):
        self.msg = [u'Could not initialize traits:']

class NoQtyError(TraitInitError):
    """ Traits do not include a qty trait. """
    def __init__(self):
        TraitInitError.__init__(self)
        self.msg.append(u'qty trait not set.')

class NoUnitError(TraitInitError):
    """ Traits do not include a unit trait. """
    def __init__(self):
        TraitInitError.__init__(self)
        self.msg.append(u'unit trait not set.')

class SetArbUnitError(Error):
    """ Errors that occur in ArbUnits.__setitem__. """
    def __init__(self, key, value):
        self.key = key
        self.value = value
        self.msg = [u'Failed to set unit %s:' % key]

class ArbUnitValueError(SetArbUnitError):
    """ Arb unit will not convert to a float. """

    def __init__(self, key, value):
        SetArbUnitError.__init__(self, key, value)
        self.msg.append('Value %s is not a number' % value)
