# Pantry, command-line nutrient analysis. Version 11
# Copyright 2007 Omari Norman. 
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE. 
# 
# Pantry's "built-in" XML validator is xmlproc.
# xmlproc is Copyright 1998-2000 by Lars Marius Garshol, Oslo, Norway. 
# 
# All Rights Reserved 
# 
# Permission to use, copy, modify, and distribute this software
# and its documentation for any purpose and without fee is hereby
# granted, provided that the above copyright notice appear in all
# copies and that modified copies are clearly marked as such. 
# 
# LARS MARIUS GARSHOL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
# SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS, IN NO EVENT SHALL LARS MARIUS GARSHOL BE LIABLE FOR
# ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
# IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
# ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
# THIS SOFTWARE. 
""" Function(s) to help when printing Unicode strings.

This module exists primarily for two reasons:

1. There are some bugs (or, "unexpected behaviors" if I'm in a
charitable mood) in Python's handling of Unicode when it is going to
stdout. See this bug:
    http://preview.tinyurl.com/23qq4d

2. The 'print' statement is very complex and, I think,
unpredictable.  (It can of course be predicted, but it is so complex
that it is not worth trying to predict what it will do.) Thus,
although there is a fix for the above bug (see the bug itself) I
find it simpler to dispense with 'print' altogether.

Python 3000 seems to be taking an encouraging approach to both these
issues: all strings will be Unicode (there will be a separate "bytes" type)
and 'print' will become a function. """

from __future__ import division
import locale
import codecs
from encodings import aliases

# see if encoding is an alias. If it is not, just leave it as is.
_pythonizedEncoding = locale.getdefaultlocale()[1].replace('-', '_').lower()
try:
    ENCODING = aliases.aliases[_pythonizedEncoding].replace('_', '-')
except KeyError:
    ENCODING = locale.getdefaultlocale()[1]

WARNING = 1
ERROR = 2

class puni(object):
    """ Prints unicode string to file using correct encoding.

    Strictly speaking the argument does not *need* to be a Unicode string,
    but if it is anything other than a Unicode string or a string that
    contains only characters in range(128), you're asking for it... 
    
    More specifically, Python will try to decode such a string first, and
    it will use the ascii codec, which will raise exceptions for a string
    that has anything outside of range(128). 
    
    If outfile is None, use sys.stdout. """
    ENCODING = ENCODING

    def __init__(self, filename = None):
        import sys
        if filename == None:
            self.outfile = sys.stdout
        else:
            self.outfile = filename

    def __call__(self, uniStr):
        toWrite = uniStr.encode(ENCODING)
        self.outfile.write(toWrite)
        return toWrite

class ErrorPrinter:
    """ Prints error messages."""
    def __init__(self, outputPuni, prog):
        """ Set up ErrorPrinter. Arguments:

        outputPuni - instance of puni to use for printing strings.
        Typically will be a puni instance hooked to stderr.

        prog - program name to use when printing error messages. """
        self.output = outputPuni
        self.prog = prog
    
    def __call__(self, message, eInst = None, type = WARNING):
        """ Prints an error message.

        message is what to print. This will be the first line
        of the error message.

        eInst is an exception instance. It must have an attribute,
        msg, which contains a list of strings to print. The strings
        will be printed, with each preceded by an indent character
        and each followed by a newline.

        type is the global constant WARNING or ERROR. Each is
        printed with an appropriate string. Warnings are non-fatal.
        Errors are fatal; sys.exit(2) is called. """

        import sys
        
        self.output('==========\n')
        if type == WARNING:
            typeStr = u'warning'
        elif type == ERROR:
            typeStr = u'error'
        self.output(u'%s: %s: %s\n' % (self.prog, typeStr, message))

        if eInst:
            for line in eInst.msg:
                self.output(u'%s\n' % line)
        self.output('==========\n')
        
        if type == ERROR:
            sys.exit(2)

def convertUni(toConvert):
    """ Convert unicode to a number. 

    Raises UnconvertableError if that is the case. 
    if toConvert is already a number, returns that."""

    import re
    
    if isinstance(toConvert, int) or isinstance(toConvert, float):
        return toConvert
    if len(toConvert) == 0:
        return None

    # see if it is a fraction or mixed number; if not, assume
    # it is just a regular number
    regex = re.compile(r'^([0-9]*\.?[0-9]* )?(([0-9]+)/([0-9]+))?$')
    match = regex.search(toConvert)
    if match:
        number = match.group(1)

        if number:
            number = number.rstrip()
            number = convertUni(number)
        else:
            number = 0

        if match.group(3):
            numerator = match.group(3)
            denominator = match.group(4)
            denominator = convertUni(denominator)
            numerator = convertUni(numerator)
            fraction = numerator / denominator
        else:
            raise UnconvertableError()
        result = number + fraction
        return result

    if '.' in toConvert:
        converter = float
    else:
        converter = int

    try:
        result = converter(toConvert)
    except ValueError:
        raise UnconvertableError()
    return result

def convertOptions(options):
    """ Convert string options in optparse to unicode. """
    
    class Holder:
        pass
    result = Holder()

    oldVars = vars(options)
    newVars = {}
    for optName, optValue in oldVars.items():
        try:
            newVars[optName] = optValue.decode(ENCODING)
        except AttributeError:
            # can't be decoded; maybe it is an iterable that
            # can be decoded
            try:
                newVars[optName] = convertList(optValue)
            except TypeError:
                # not an iterable either; just add it to the result
                newVars[optName] = optValue

    for optName, optValue in newVars.items():
        setattr(result, optName, optValue)

    return result

def convertList(toConvert):
    """ Returns list or tuple with items converted to unicode.

    Only attempts to convert items that are strings or unicode. """

    result = []
    for item in toConvert:
        try:
            newItem = item.decode(ENCODING)
        except (AttributeError, TypeError):
            result.append(item)
        else:
            result.append(newItem)
    return result

class Error(Exception):
    pass

class UnconvertableError(Error):
    """ convertUni failed because arg can't be converted. """
    pass
